
# Standard packages/modules
import requests
from collections import namedtuple

# installed packages/modules
from bs4 import BeautifulSoup

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.common.exceptions import NoSuchElementException

# my pagkages/modules

class SiriusScraper(object):

	"""docstring for Scraper"""
	
	# TODO: Maybe have named tuple to manage the channeldata fromt he table

	URL_FOR_COOKIE = 'http://www.siriusxm.com'
	URI_ALL_CHANNELS = 'http://www.siriusxm.com/channellineup/'
	CHANNEL_TABLE_ID = 'packageChannelsTable'

	CHANNEL_INTERNAL_ID_VAR = 'SXM.ChannelContentID = '

	@classmethod
	def get_channel_id(cls,url):

		page = requests.get(url).text
		soup = BeautifulSoup(page,'html.parser')

		scripts = soup.findAll('script')
		for script in scripts:
			scriptString = script.string
			if scriptString:
				found = scriptString.find(cls.CHANNEL_INTERNAL_ID_VAR)
				if found > -1:
					start = found + len(cls.CHANNEL_INTERNAL_ID_VAR)
					end = scriptString.find(";",start)
					return scriptString[start:end][1:-1]
		return None

	def __init__(self):
		super(SiriusScraper, self).__init__()

		# TODO: might want to use PhantomJs
		self.browser = webdriver.Firefox()

		# open browser to bogus url for website
		self.browser.get(type(self).URL_FOR_COOKIE)

		# set the platform choice cookie on bogus page
		cookie = {'name' : "sxm_platform", 'value' : 'sirius'}
		self.browser.add_cookie(cookie)
		
		# now nav to the all channels page for scraping
		self.browser.get(type(self).URI_ALL_CHANNELS)

		self._channel_table = self.browser.find_element_by_id(type(self).CHANNEL_TABLE_ID)

	def get_channel_data(
		self,
		channel = None,
		channel_name = None):

		if (not channel and
			not channel_name):
			raise ValueError('One of channel,channel_name is required.')

		if channel:
			source = self._get_channel_page_source(
				channel_id_type = 'channelnum',
				channel_id = channel)
		elif channel_name:
			source = self._get_channel_page_source(
				channel_id_type = 'channelname',
				channel_id = channel_name)
			
		# TODO: Raise exception
		if source:
			return self._get_channel_data_from_page_source(source)
		else:
			return None

	def get_channels_data(self):

		# temp unsorted list of Channel object instances
		channels_data = []

		for row in self._channel_table.find_elements_by_tag_name('tr'):

			cell_values = self._get_channel_data_from_page_source(row)

			if cell_values:
				channels_data.append(cell_values)

		# need to cast to int otherwise we get order like 1,10,2,20,3 
		# instead of 1,2,3,10,20
		return sorted(channels_data, key=lambda c: int(c[0]))

	def _get_channel_page_source(
		self,
		channel_id_type,
		channel_id):

		# TODO: Maube catch exception selenium.common.exceptions.NoSuchElementException
		# if the item doesnt exist and raise my own

		if channel_id_type == 'channelnum':
			return self._channel_table.find_element_by_xpath(
				'//span[@class="channelnum" and contains(text(), "{0}")]'.format(
					str(channel_id))).find_element_by_xpath('../../..')
		elif channel_id_type == 'channelname':

			channel_name_cells = self._channel_table.find_elements_by_class_name('channelname')

			if channel_name_cells:

				for channel_name_cell in channel_name_cells:

					try:
						link = channel_name_cell.find_element_by_link_text(
							channel_id).find_element_by_xpath(
								'../..')

						return link
					except NoSuchElementException:
						pass
		return None

	def _get_channel_data_from_page_source(self,page_source):

		channel_data = None
		cells = page_source.find_elements_by_tag_name('td')

		if cells:
			# TODO: make sure we make use of the named tuple here instead
			channel_number = 0
			channel_name = 1
			channel_category = 2
			channel_genre = 3
			channel_desc = 4
			url = cells[channel_number].find_element_by_tag_name('a').get_attribute('href')
			channel_data = (int(cells[channel_number].text.encode('ascii','ignore')),
				 			type(self).get_channel_id(url),
							cells[channel_name].text.encode('ascii','ignore'),
							url,
							cells[channel_category].text.encode('ascii','ignore'),
							cells[channel_genre].text.encode('ascii','ignore'),
							cells[channel_desc].text.encode('ascii','ignore'))

		return channel_data

	def __del__(self):

		self.browser.quit()


# standard/installed
import requests
from functools import reduce
from operator import getitem
from datetime import datetime, timedelta
from time import time


class SiriusCurrentlyPlaying(object):

	JSON_URI_PREFIX = 'http://www.siriusxm.com/metadata/pdt/en-us/json/channels/'
	JSON_MSG_SUCCESS_CODES = [100,200]

	JSON_ROOT_MSG_DATA_KEYS = {
								'messages':['channelMetadataResponse','messages'],
								'song':['channelMetadataResponse','metaData','currentEvent'],
		}

	JSON_MSG_DATA_KEYS = {
							'album':['song','album','name'],
							'artist':['artists','name'],
							'code':['code'],
							'song':['song','name'],
	}

	def __init__(self,channel_id):
		super(SiriusCurrentlyPlaying,self).__init__()

		self.id = channel_id
		self.__last_updated = None
		self.data = self.__get_currently_playing()

	def __get_currently_playing(self):

		# TODO make sure this is the correct way to get UTC time
		timenow = (datetime.utcnow() - timedelta(minutes=1)).strftime('%m-%d-%H:%M:00')
		urlSuffix = '/timestamp/'+timenow
		url = type(self).JSON_URI_PREFIX + self.id + urlSuffix

		self.last_updated = time()
		return requests.get(url).json()

	def update(self):

		# dont want to overload XM's servers, only ping every minute
		# they tend to update currently playing on their site as well.
 		if not ((time() - self.last_updated) < 60):
			self.data = self.__get_currently_playing()

	@property
	def artist_name(self):
		keys = type(self).JSON_ROOT_MSG_DATA_KEYS['song'] + type(self).JSON_MSG_DATA_KEYS['artist']
		return self.currently_playing_item(keys)

	@property
	def song_name(self):
		keys = type(self).JSON_ROOT_MSG_DATA_KEYS['song'] + type(self).JSON_MSG_DATA_KEYS['song']
		return self.currently_playing_item(keys)
	
	@property
	def album(self):
		keys = type(self).JSON_ROOT_MSG_DATA_KEYS['song'] + type(self).JSON_MSG_DATA_KEYS['album']
		return self.currently_playing_item(keys)

	# TODO: change this to __getitem__ or overload/subclass it somehow
	# or catch the key exception and raise my own
	def currently_playing_item(self,keys):

		# TODO: some items have issues on their return values
		# DR DRE: albumn 2001 song: forgot about dre has issues since its an int 2001
		# other international channels have issues with unicode values
		msg_code_keys = type(self).JSON_ROOT_MSG_DATA_KEYS['messages'] + type(self).JSON_MSG_DATA_KEYS['code']

		message = reduce(getitem,msg_code_keys,self.data) 

		if message in type(self).JSON_MSG_SUCCESS_CODES:
			return reduce(getitem,keys,self.data)
		else:
			print(message)
			return None


# standard

# installed

# my
from sirius_playing import SiriusCurrentlyPlaying

class SiriusChannel(object):

	"""docstring for Channel"""
	def __init__(
			self,
			channel,
			id,
			name,
			url,
			category= None,
			genre = None,
			description = None,
			*args,
			**kwargs):
		# PY3 super().__init__(*args,**kwargs)
		super(SiriusChannel,self).__init__()
		
		self.channel = channel
		self.name = name
		self.url = url
		self.id = id

		self.category = category
		self.genre = genre
		self.description = description

		self.currently_playing = None

	def PrintMe(self):

		print('Channel:' + str(self.channel))
		print('\t' + 'ID: ' + str(self.id))
		print('\t' + 'Name: ' + str(self.name))
		print('\t' + 'URL: ' + str(self.url))
		print('\t' + 'Category: ' + str(self.category))
		print('\t' + 'Genre: ' + str(self.genre))
		print('\t' + 'Description: ' + str(self.description))

	def get_currently_playing(self):

		if not self.currently_playing:
			self.currently_playing = SiriusCurrentlyPlaying(self.id)
		else:
			self.currently_playing.update()

		return self.currently_playing
		






# Standard packages/modules
from collections import namedtuple

# installed packages/modules

# my pagkages/modules
from defaults import *

class PySoSirius(object):

	channel_data_wrapper = namedtuple(
							'wrap_channel_data',
							['channel','id','name','url','category','genre','description']
							)

	@classmethod
	def wrap_channel_data(cls,channel_data):

		return cls.channel_data_wrapper(*channel_data)

	"""docstring for PSS"""
	def __init__(self,*args,**kwargs):
		# PY3 super().__init__(*args,**kwargs)
		super(PySoSirius,self).__init__(*args,**kwargs)

		self.channels = []

		if 'database' in kwargs.keys():
			self.database = kwargs['database']
		else:
			self.database = DATABASE
		
		if 'genres' in kwargs.keys():
			self.genres = kwargs['genres']
		else:
			self.genres = LIMIT_TO_GENRES

		if 'categories' in kwargs.keys():
			self.categories = kwargs['categories']
		else:
			self.categories = LIMIT_TO_CATEGORIES

		if 'channels' in kwargs.keys():
			self.channels = kwarg['channels']
		else:
			self.channels = LIMIT_TO_CHANNELS
	'''
		# Handle DB level init stuff
		if self.database:
			self.db_connex = sqlite3.connect(self.database)
			self.db_conn_cursor = self.db_connex.cursor()
		# TODO: Need to have a DB Init to build local tables
		# for shit

	def __del__(self,**kwargs):
		super().__del__(kwargs)

		if self.database:
			self.db_conn_cursor.close()
	'''
	
	def get_channels(self):

		from database.manager import SiriusManager
		from sirius_channel import SiriusChannel

		channels = []
		db_manager = SiriusManager()

		# TODO this For could probably be a lambda or map
		# some sort of one liner
		channel_rows = db_manager.get_channel_rows()
		for channel_row in channel_rows:

			channels.append(SiriusChannel(*channel_row))

		return channels


	def get_channel(self,**kwargs):

		from database.manager import SiriusManager
		from sirius_channel import SiriusChannel

		db_manager = SiriusManager()

		channel_row = db_manager.get_channel_row(**kwargs)
		
		return SiriusChannel(*channel_row)



from setuptools import setup, find_packages

'''
When creating a new version 

----- Steps -------
o Create new version number
o Delete existing Dist folder (.tar and whl)
o python setup.py sdist
o python setup.py bdist_wheel (Python 2.7)
o python setup.py register -r pypi (located in .pypirc file in home)
o twine upload dist/*

----- Links -----
https://wiki.python.org/moin/TestPyPI
https://testpypi.python.org/
'''


setup(
	name='pysosirius',
	version='0.0.3',
	description='Unofficial Sirius XM API',
	url='https://github.com/T-Santos/PySoSirius',
	author='Tyler Santos',
	author_email='1tsantos7+pysosirius@Gmail.com',
	license='MIT',
	packages=['pysosirius','pysosirius.database','pysosirius.web'],
	#packages = find_packages(),
	# Project uses BeautifulSoup, requests and selenium to
	# parse webpages for new info not stored in the local db
	install_requires=['bs4','requests','selenium'],
	package_data={
	    'pysosirius.database': ['PySoSirius.db'],
	},
	keywords="Sirius SiriusXM XM Music",	
	classifiers=[
		# How mature is this project? Common values are
		'Development Status :: 1 - Planning',
		#Development Status :: 2 - Pre-Alpha,
		#Development Status :: 3 - Alpha,
		#Development Status :: 4 - Beta,
		#Development Status :: 5 - Production/Stable,
		#Development Status :: 6 - Mature,
		#Development Status :: 7 - Inactive,

		# Indicate who your project is intended for
		'Intended Audience :: Developers',

		# Pick your license as you wish (should match "license" above)
		 'License :: OSI Approved :: MIT License',

		# Specify the Python versions you support here. In particular, ensure
		# that you indicate whether you support Python 2, Python 3 or both.
		'Programming Language :: Python :: 2.7'
	],
	zip_safe=True)


# Standard
import sqlite3
import os

# installed

# mine

class SiriusManager(object):

	DATABASE_FILENAME = 'PySoSirius.db'

	"""docstring for SiriusManager"""
	def __init__(self):
		super(SiriusManager, self).__init__()

		self.database_path = os.path.join(os.path.dirname(__file__), type(self).DATABASE_FILENAME)

		self.connection = sqlite3.connect(self.database_path)
		self.cursor = self.connection.cursor()

		self._create_channel_table()

		def __del__(self):

			self.connection.close()
	
	def _create_channel_table(self):

		# Create table
		try:
			self.cursor.execute(
				'''CREATE TABLE channels
					(channel INTEGER PRIMARY KEY NOT NULL,
					 id TEXT NOT NULL,
					 name TEXT NOT NULL,
					 url TEXT NOT NULL,
					 category TEXT NOT NULL,
					 genre TEXT NOT NULL,
					 description TEXT)''')
		except sqlite3.OperationalError:
			pass	

	def get_channel_rows(self):

		# TODO: look into yield or iterator of sorts
		self.cursor.execute('SELECT * FROM channels')
		return self.cursor.fetchall()
		
	def get_channel_row(
			self,
			channel = None,
			id = None,
			name = None,
			url = None):

		if (not channel and
			not id and
			not name and
			not url):
			raise ValueError('One of channel, id, name, url is required.')

		if channel:
			self.cursor.execute('SELECT * FROM channels WHERE channel=?',(channel,))
		elif id:
			self.cursor.execute('SELECT * FROM channels WHERE id=?',(id,))
		elif name:
			self.cursor.execute('SELECT * FROM channels WHERE name=?',(name,))
		elif url:
			self.cursor.execute('SELECT * FROM channels WHERE url=?',(url,))

		return self.cursor.fetchone()

	def set_channel_row(self,data):

		# TODO: Raise own exception 
		try:
			self.cursor.execute('INSERT INTO channels VALUES (?,?,?,?,?,?,?)',data)
		except sqlite3.IntegrityError:
			pass

	def save(self):
		self.connection.commit()

	def update_database(self):

		# given a list of channels, see if the rows for those 
		# channels need to be updated or added to the db
		pass
